# Honeybee: A Plugin for Environmental Analysis (GPL)
# This file is part of Honeybee.
#
# Copyright (c) 2021, Ladybug Tools.
# You should have received a copy of the GNU General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""
Create a list of modifiers that can be used to edit or create a ModifierSet object.
-

    Args:
        _exterior_shade_: A modifier object for exterior shades (or text for
            the identifier of the modifier within the library).
        _interior_shade_: A modifier object for interior shades (or text for
            the identifier of the modifier within the library).
    
    Returns:
        shade_set: A list of shade modifiers that can be used to edit or create
            a ModifierSet object.
"""

ghenv.Component.Name = 'HB Shade Modifier Subset'
ghenv.Component.NickName = 'ShadeSubset'
ghenv.Component.Message = '1.3.0'
ghenv.Component.Category = 'HB-Radiance'
ghenv.Component.SubCategory = '1 :: Modifiers'
ghenv.Component.AdditionalHelpFromDocStrings = '4'

try:  # import honeybee_radiance dependencies
    from honeybee_radiance.modifier import Modifier
    from honeybee_radiance.lib.modifiers import modifier_by_identifier
except ImportError as e:
    raise ImportError('\nFailed to import honeybee_radiance:\n\t{}'.format(e))


def check_mod(mod, input_name):
    """Get an Modifier from the library if it's a string."""
    if isinstance(mod, str):
        return modifier_by_identifier(mod)
    else:
        assert isinstance(mod, Modifier), \
            'Expected Modifier for {}. Got {}'.format(input_name, type(mod))
    return mod


# go through each input modifier
if _exterior_shade_ is not None:
    _exterior_shade_ = check_mod(_exterior_shade_, '_exterior_shade_')
if _interior_shade_ is not None:
    _interior_shade_ = check_mod(_interior_shade_, '_interior_shade_')

# return the final list from the component
shade_set = [_exterior_shade_, _interior_shade_]



# Honeybee: A Plugin for Environmental Analysis (GPL)
# This file is part of Honeybee.
#
# Copyright (c) 2021, Ladybug Tools.
# You should have received a copy of the GNU General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""
Create an opaque radiance modifier from red, green, and blue reflectances.
-

    Args:
        _name_: Text to set the name for the modifier and to be incorporated into
            a unique modifier identifier.
        _r_ref: A number between 0 and 1 for the red reflectance.
        _g_ref: A number between 0 and 1 for the green reflectance.
        _b_ref: A number between 0 and 1 for the blue reflectance.
        _spec_: A number between 0 and 1 for the fraction of specularity. Specularity
            fractions greater than 0.1 are not common in non-metallic
            materials. (Default: 0).
        _rough_: Roughness is specified as the rms slope of surface facets. A value
            of 0 corresponds to a perfectly smooth surface, and a value of 1 would be
            a very rough surface. Roughness values greater than 0.2 are not very
            realistic. (Default: 0).

    Returns:
        modifier: An opaque modifier that can be assigned to a Honeybee geometry
            or Modifier Sets.
"""

ghenv.Component.Name = 'HB Opaque Modifier 3'
ghenv.Component.NickName = 'OpaqueMod3'
ghenv.Component.Message = '1.3.0'
ghenv.Component.Category = 'HB-Radiance'
ghenv.Component.SubCategory = '1 :: Modifiers'
ghenv.Component.AdditionalHelpFromDocStrings = '0'

try:  # import the core honeybee dependencies
    from honeybee.typing import clean_and_id_rad_string, clean_rad_string
except ImportError as e:
    raise ImportError('\nFailed to import honeybee:\n\t{}'.format(e))

try:  # import the honeybee-radiance dependencies
    from honeybee_radiance.modifier.material import Plastic
except ImportError as e:
    raise ImportError('\nFailed to import honeybee_radiance:\n\t{}'.format(e))

try:  # import ladybug_rhino dependencies
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component):
    # set the default modifier properties
    _spec_ = 0.0 if _spec_ is None else _spec_
    _rough_ = 0.0 if _rough_ is None else _rough_
    name = clean_and_id_rad_string('OpaqueMaterial') if _name_ is None else \
        clean_rad_string(_name_)

    # create the modifier
    modifier = Plastic(name, _r_ref, _g_ref, _b_ref, _spec_, _rough_)
    if _name_ is not None:
        modifier.display_name = _name_



# Honeybee: A Plugin for Environmental Analysis (GPL)
# This file is part of Honeybee.
#
# Copyright (c) 2021, Ladybug Tools.
# You should have received a copy of the GNU General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>


"""
Deconstruct a modifier set into its constituient interior modifiers.
-

    Args:
        _mod_set: A modifier set to be deconstructed. This can also be
            text for a modifier set to be looked up in the modifier
            set library.

    Returns:
        interior_wall: A modifier object for the set's interior walls.
        ceiling: A modifier object for the set's interior roofs.
        interior_floor: A modifier object for the set's interior floors.
        interior_window: A modifier object for all apertures with a Surface
            boundary condition.
        interior_door: A modifier object for all opaque doors with a Surface
            boundary condition.
        int_glass_door: A modifier object for all glass doors with a Surface
            boundary condition.
        interior_shade: A modifier object for all interior shades.
"""

ghenv.Component.Name = 'HB Deconstruct ModifierSet Interior'
ghenv.Component.NickName = 'DecnstrConstrSetInt'
ghenv.Component.Message = '1.3.0'
ghenv.Component.Category = 'HB-Radiance'
ghenv.Component.SubCategory = '1 :: Modifiers'
ghenv.Component.AdditionalHelpFromDocStrings = '5'


try:  # import the honeybee-radiance dependencies
    from honeybee_radiance.modifierset import ModifierSet
    from honeybee_radiance.lib.modifiersets import modifier_set_by_identifier
except ImportError as e:
    raise ImportError('\nFailed to import honeybee_radiance:\n\t{}'.format(e))

try:  # import ladybug_rhino dependencies
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component):
    # check the input
    if isinstance(_mod_set, str):
        _mod_set = modifier_set_by_identifier(_mod_set)
    else:
        assert isinstance(_mod_set, ModifierSet), \
            'Expected ModifierSet. Got {}.'.format(type(_mod_set))

    interior_wall = _mod_set.wall_set.interior_modifier
    ceiling = _mod_set.roof_ceiling_set.interior_modifier
    interior_floor = _mod_set.floor_set.interior_modifier
    interior_window = _mod_set.aperture_set.interior_modifier
    interior_door = _mod_set.door_set.interior_modifier
    int_glass_door = _mod_set.door_set.interior_glass_modifier
    interior_shade = _mod_set.shade_set.interior_modifier



# Honeybee: A Plugin for Environmental Analysis (GPL)
# This file is part of Honeybee.
#
# Copyright (c) 2021, Ladybug Tools.
# You should have received a copy of the GNU General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""
Combine Honeybee Shades into a single dynamic group. Shades that are a part
of the same dynamic group will have their states change in unison. If an shade
has no dynamic group, it is assumed to be static.
-
This component can also be used to combine shades that already have states
assigned to them into one group since existing states are not overwritten if
nothing is connected to states_. In this case, the total number of states in
the dynamic group is equal to that of the object with the highest number of
states. After a dynamic shade with fewer states than that of it's dynamic
group has hit its highest state, it remains in that state as the other dynamic
shades continue to change.
-

    Args:
        _shades: A list of Honeybee Shades to be grouped together into a
            single dynamic group.
        _name_: Text to be incorporated into a unique identifier for the dynamic
            Shade group. If the name is not provided, a random name will be assigned.
        states_: An optional list of Honeybee State objects ordered based on
            how they will be switched on. The first state is the default state
            and, typically, higher states are more shaded. If the objects in the
            group have no states, the modifiers already assigned the shades
            will be used for all states.
    
    Returns:
        group_shds: Honeybee shades that are a part of the same dynamic group.
            These can be used directly in radiance simulations or can be added
            to Honeybee faces and rooms.
"""

ghenv.Component.Name = 'HB Dynamic Shade Group'
ghenv.Component.NickName = 'ShadeGroup'
ghenv.Component.Message = '1.3.0'
ghenv.Component.Category = 'HB-Radiance'
ghenv.Component.SubCategory = '0 :: Basic Properties'
ghenv.Component.AdditionalHelpFromDocStrings = '2'

import uuid

try:  # import the core honeybee dependencies
    from honeybee.shade import Shade
    from honeybee.typing import clean_and_id_rad_string, clean_rad_string
except ImportError as e:
    raise ImportError('\nFailed to import honeybee:\n\t{}'.format(e))

try:  # import the honeybee-radiance dependencies
    from honeybee_radiance.dynamic import RadianceShadeState
except ImportError as e:
    raise ImportError('\nFailed to import honeybee_radiance:\n\t{}'.format(e))

try:  # import the ladybug_rhino dependencies
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component):
    # check and duplicate the input objects
    group_shds = []
    for shd in _shades:
        assert isinstance(shd, Shade), 'Expected Shade ' \
            'for dynamic shade group. Got {}.'.format(type(shd))
        group_shds.append(shd.duplicate())

    # set the name of the dynamic group
    name = clean_and_id_rad_string('ShadeGroup') if _name_ is None else clean_rad_string(_name_)
    for shd in group_shds:
        shd.properties.radiance.dynamic_group_identifier = name

    # assign any states if they are connected
    if len(states_) != 0:
        # convert the sub-face states to shade states
        shd_states = [RadianceShadeState(st.modifier, st.shades) for st in states_]
        # assign states (including shades) to the first shade
        group_shds[0].properties.radiance.states = [state.duplicate() for state in shd_states]

        # remove shades from following shades to ensure they aren't double-counted
        states_wo_shades = []
        for state in shd_states:
            new_state = state.duplicate()
            new_state.remove_shades()
            states_wo_shades.append(new_state)
        for shd in group_shds[1:]:
            shd.properties.radiance.states = \
                [state.duplicate() for state in states_wo_shades]



# This file is part of Honeybee.
#
# Copyright (c) 2021, Ladybug Tools.
# You should have received a copy of the GNU General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""
Perform glare post-processing on a hemisphical fisheye HDR image file.
_
Glare post-processing includes calcuating Daylight Glare Probability (DGP) as
well as other glare indexes (DGI, UGR, VCP, CGI, UDP).
_
This component is using the `evalglare` function for glare calculations., which
is developed by J. Wienold at Fraunhofer ISE. More information on evalglare
can be found here:
https://www.radiance-online.org/learning/documentation/manual-pages/pdfs/evalglare.pdf/view
_
For more information about the metrics used to evaluate glare, see here:
http://web.mit.edu/tito_/www/Projects/Glare/GlareRecommendationsForPractice.html

-
    Args:
        _hdr: Path to a hemisphical fisheye High Dynamic Range (HDR) image file. This can be
            obtained from the "HB Point-In-Time View-Based" recipe component. Due
            to runtime reasons of the evalglare code, the input HDR image should
            be smaller than 1500 x 1500 pixels. The recommended size is 1000 x 1000
            pixels, the minimum recommended size is 800 x 800 pixels.
        task_pos_: An optional task position as a 2D point or string formatted as "X, Y".
            The X and Y coordinates of this point must be numbers between 0 and 1
            and correspond to fraction of the image width and height where the
            task position lies. If no task position is provided, the glare will
            be valuated for the entire scene of the image.
        task_angle_: An number between 0 and 180 for the task position opening angle
            in degrees. This angle indicates how widely the peripheral vision
            is engaged for the task. (Default: 30).
        hide_task_: Boolean to note whether the task area should be hidden in the
            output check_hdr.

    Returns:
        DGP: Daylight Glare Probability (DGP) as a number between 0 and 1. The DGP
            describes the fraction of persons disturbed by glare, where 0 is no
            one disturbed and 1 is everyone. Values lower than 0.2 are out of the
            range of the user assessment tests, where the program is based on and
            should be interpreted carefully.
        category: Text for the category of glare discomfort. It will be one of the
            following.
                * Imperceptible Glare [0.35 > DGP]
                * Perceptible Glare [0.4 > DGP >= 0.35]
                * Disturbing Glare [0.45 > DGP >= 0.4]
                * Intolerable Glare [DGP >= 0.45] 
        glare_indices: A list of various glare indices ordered as follows.
                * Daylight Glare Index (DGI)
                * Unified Glare Rating (UGR)
                * Visual Comfort Probability (VCP)
                * CIE Glare Index (CGI)
                * Veiling Luminance (Lveil)
        check_hdr: Path to a HDR image produced from the glare study. The image will
            use randomly-assigned colors to indicate different sources of glare
            in the image. It will also show a circular region for the task area
            unless hide_task_ has been set to True.
"""

ghenv.Component.Name = 'HB Glare Postprocess'
ghenv.Component.NickName = 'Glare'
ghenv.Component.Message = '1.3.1'
ghenv.Component.Category = 'HB-Radiance'
ghenv.Component.SubCategory = '4 :: Results'
ghenv.Component.AdditionalHelpFromDocStrings = '3'

import os
import subprocess
import math

try:  # import honeybee_radiance dependencies
    from honeybee_radiance.config import folders as rad_folders
except ImportError as e:
    raise ImportError('\nFailed to import honeybee_radiance:\n\t{}'.format(e))

try:  # import ladybug_rhino dependencies
    from ladybug_rhino.grasshopper import all_required_inputs, give_warning
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))

# check the Radiance date of the installed radiance
try:  # import lbt_recipes dependencies
    from lbt_recipes.version import check_radiance_date
except ImportError as e:
    raise ImportError('\nFailed to import lbt_recipes:\n\t{}'.format(e))
check_radiance_date()


def check_hdr_luminance_and_fisheye(hdr_path):
    """Check that a given HDR file is a fisheye image for visible luminance.

    A ValueError is raised if the image is not for luminance or if the image is
    not clearly a hemispheical fisheye.

    Args:
        hdr_path: The path to an HDR image file.
    """
    msg = 'Connected _hdr image must be for luminance. Got "{}".'
    projection = '-vth'
    with open(hdr_path, 'r') as hdr_file:
        for lineCount, line in enumerate(hdr_file):
            if lineCount < 200:
                low_line = line.strip().lower()
                if low_line.startswith('rpict'):
                    if line.find('_irradiance.vf') > -1:
                        raise ValueError(msg.format('irradiance'))
                    if line.find('_radiance.vf') > -1:
                        raise ValueError(msg.format('radiance'))
                    if line.find('-i') > -1 and not line.find('-i-') > -1:
                        raise ValueError(msg.format('illuminance'))
                elif low_line.startswith('view='):
                    if line.find('-vth') > -1:
                        projection = '-vth'
                    elif line.find('-vta') > -1:
                        projection = '-vta'
                    else:
                        raise ValueError(
                            'Connected _hdr image is not a fisheye projection.\n'
                            'Make sure the view type of the image is 1(h) or 4(a).')
                elif 'pcond -h' in low_line:
                    raise ValueError(
                        'Connected _hdr image has had the exposure adjusted on it.\n'
                        'Make sure adj_expos_ has been set to False in previous steps.')
            else:  # no need to check the rest of the document
                break
    return projection

def check_hdr_dimensions(hdr_path):
    """Check that a given HDR file has dimensions suitable for evalglare.

    A warning is raised if the image is not 1000x1000 pixels and a ValueError is
    raised if the image is completely outside the accptable ragne from 800x800
    to 1500x1500 pixels.

    Args:
        hdr_path: The path to an HDR image file.
    """
    # get the path the the getinfo command
    getinfo_exe = os.path.join(rad_folders.radbin_path, 'getinfo.exe') if \
        os.name == 'nt' else os.path.join(rad_folders.radbin_path, 'getinfo')

    # run the getinfo command in a manner that lets us obtain the result
    cmds = [getinfo_exe, '-d', hdr_path]
    use_shell = True if os.name == 'nt' else False
    process = subprocess.Popen(cmds, stdout=subprocess.PIPE, shell=use_shell)
    stdout = process.communicate()
    img_dim = stdout[0]

    # check the X and Y dimensions of the image
    x = int(img_dim.split(' ')[-1].strip())
    y = int(img_dim.split(' ')[-3].strip())
    msg = 'Recommended _hdr image dimensions for glare analysis should be \n' \
        '{} {} x {} pixels. Got {} x {}.'
    if x < 800 or y < 800:
        give_warning(ghenv.Component, msg.format('at least', 800, 800, x, y))
    elif x > 1500 or y > 1500:
        give_warning(ghenv.Component, msg.format('no greater than', 1500, 1500, x, y))
    return x, y


def dgp_comfort_category(dgp):
    """Get text for the glare comfort category given a DGP value."""
    if dgp < 0.35:
        return 'Imperceptible Glare'
    elif dgp < 0.40:
        return 'Perceptible Glare'
    elif dgp < 0.45:
        return 'Disturbing Glare'
    else:
        return 'Intolerable Glare'


if all_required_inputs(ghenv.Component):
    # check the input image to ensure it meets the criteria
    projection = check_hdr_luminance_and_fisheye(_hdr)
    width, height = check_hdr_dimensions(_hdr)

    # get the path the the evalglare command and setup the check image argument
    evalglare_exe = os.path.join(rad_folders.radbin_path, 'evalglare.exe') if \
        os.name == 'nt' else os.path.join(rad_folders.radbin_path, 'evalglare')
    img_dir = os.path.dirname(_hdr)
    input_image = os.path.basename(_hdr)
    new_image = input_image.lower().replace('.hdr', '_check.HDR')
    check_hdr = os.path.join(img_dir, new_image)
    cmds = [evalglare_exe, '-c', check_hdr]

    # since pcomp is used to merge images, the input usually doesn't have view information
    # add default view information for hemispheical fish-eye camera
    cmds.extend([projection, '-vv', '180', '-vh', '180'])

    # process the task position and add the input HDR
    if task_pos_:
        uv_pt = [float(val) for val in task_pos_.split(',')]
        assert 0 <= uv_pt[0] <= 1 and 0 <= uv_pt[1] <= 1, 'Task position X and Y ' \
            'coordinates must be between 0 and 1.'
        angle = math.radians(task_angle_) if task_angle_ is not None else math.radians(30)
        task_opt = '-t' if hide_task_ else '-T'
        cmds.extend(
            [task_opt, str(int(uv_pt[0] * width)), str(int(uv_pt[1] * height)), str(angle)])
    cmds.append(_hdr)

    # run the evalglare command in a manner that lets us obtain the stdout result
    use_shell = True if os.name == 'nt' else False
    process = subprocess.Popen(cmds, stdout=subprocess.PIPE, shell=use_shell)
    stdout = process.communicate()

    # process the stdout result into the component outputs
    glare_result = stdout[0].split(':')[-1].strip()
    glare_indices = [float(val) for val in glare_result.split(' ')]
    DGP = glare_indices.pop(0)
    category = dgp_comfort_category(DGP)



# Honeybee: A Plugin for Environmental Analysis (GPL)
# This file is part of Honeybee.
#
# Copyright (c) 2021, Ladybug Tools.
# You should have received a copy of the GNU General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""
Create a Wea object from an EPW file.
-

    Args:
        _epw_file: Full path to an .epw weather file.
        hoys_: An optional list of hours of the year (numbers from 0 to 8759) for
            which the Wea will be filtered. HOYs can be generated from the
            "LB Analysis Period" component or they can be obtained through
            other means like analysis of the values in an occupancy schedule.
            By default, the Wea will be generated for the whole year.
        timestep_: An integer representing the timestep with which to make the 
            WEA object.  Default is set to 1 for 1 step per hour of the year.

    Returns:
        wea: A wea object from epw file.
"""

ghenv.Component.Name = 'HB Wea From EPW'
ghenv.Component.NickName = 'Wea'
ghenv.Component.Message = '1.3.0'
ghenv.Component.Category = 'HB-Radiance'
ghenv.Component.SubCategory = '2 :: Light Sources'
ghenv.Component.AdditionalHelpFromDocStrings = '2'

try:
    from ladybug.wea import Wea
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))

try:  # import ladybug_rhino dependencies
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component):
    timestep_ = 1 if timestep_ is None else timestep_
    wea = Wea.from_epw_file(_epw_file, timestep_)
    if len(hoys_) != 0:
        wea = wea.filter_by_hoys(hoys_)



# Honeybee: A Plugin for Environmental Analysis (GPL)
# This file is part of Honeybee.
#
# Copyright (c) 2021, Ladybug Tools.
# You should have received a copy of the GNU General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""
Create an glass radiance modifier from a single transmittance.
-

    Args:
        _name_: Text to set the name for the modifier and to be incorporated into
            a unique modifier identifier.
        _trans: A number between 0 and 1 for the glass modifier transmittance.
            This transmittance will be the same for the red, green and blue channels.
        _refract_: Index of refraction. Typical values are 1.52 for float
            glass and 1.4 for ETFE. If None, Radiance will default to using 1.52
            for glass (Default: None).

    Returns:
        modifier: A glass modifier that can be assigned to a Honeybee geometry or
            Modifier Sets.
"""

ghenv.Component.Name = 'HB Glass Modifier'
ghenv.Component.NickName = 'GlassMod'
ghenv.Component.Message = '1.3.0'
ghenv.Component.Category = 'HB-Radiance'
ghenv.Component.SubCategory = '1 :: Modifiers'
ghenv.Component.AdditionalHelpFromDocStrings = '2'

try:  # import the core honeybee dependencies
    from honeybee.typing import clean_and_id_rad_string, clean_rad_string
except ImportError as e:
    raise ImportError('\nFailed to import honeybee:\n\t{}'.format(e))

try:  # import the honeybee-radiance dependencies
    from honeybee_radiance.modifier.material import Glass
except ImportError as e:
    raise ImportError('\nFailed to import honeybee_radiance:\n\t{}'.format(e))

try:  # import ladybug_rhino dependencies
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component):
    name = clean_and_id_rad_string('GlassMaterial') if _name_ is None else \
        clean_rad_string(_name_)

    # create the modifier
    modifier = Glass.from_single_transmittance(name, _trans, _refract_)
    if _name_ is not None:
        modifier.display_name = _name_



# Honeybee: A Plugin for Environmental Analysis (GPL)
# This file is part of Honeybee.
#
# Copyright (c) 2021, Ladybug Tools.
# You should have received a copy of the GNU General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""
Run a cumulative radiation study for a Honeybee model.
_
This recipe calculates cumulative radiation (kWh/m2) and average irradiance (W/m2)
over the time period of a specified Wea.
_
The fundamental calculation of this recipe is the same as that of the "LB Incident
Radiation" component except that this recipe uses Radiance and can therefore
account for ambient reflections. Like LB Incident Radiation, the direct sun in this
recipe is diffused between several sky patches and so the precise line between shadow
and sun for each hour is blurred. This approximation is acceptable for studies
where one is only concerned about the average/total conditions over time and the
timestep-by-timestep irradiance values do not need to be exact. For accurate
modeling of direct irradiance on a timestep-by-timestep basis, see the "HB Annual
Irradiance" recipe.

-
    Args:
        _model: A Honeybee Model for which Cumulative Radiation will be simulated.
            Note that this model should have grids assigned to it.
        _wea: A Wea object produced from the Wea components that are under the Light
            Sources tab. This can also be the path to a .wea or a .epw file.
        _timestep_: An integer for the timestep of the inpput _wea. (Default: 1)
        _sky_density_: An integer for the number of times that that the original Tregenza
            sky patches are subdivided. 1 indicates that 145 patches are used
            to describe the sky hemisphere, 2 indicates that 577 patches describe
            the hemisphere, and each successive value will roughly quadruple the
            number of patches used. Setting this to a high value will result in
            a more accurate analysis but will take longer to run. (Default: 1).
        north_: A number between -360 and 360 for the counterclockwise difference
            between the North and the positive Y-axis in degrees. This can
            also be Vector for the direction to North. (Default: 0).
        grid_filter_: Text for a grid identifer or a pattern to filter the sensor grids of
            the model that are simulated. For instance, first_floor_* will simulate
            only the sensor grids that have an identifier that starts with
            first_floor_. By default, all grids in the model will be simulated.
        sensor_count_: Integer for the maximum number of sensor grid points per
            parallel execution. (Default: 200).
        radiance_par_: Text for the radiance parameters to be used for ray
            tracing. (Default: -ab 2 -ad 5000 -lw 2e-05).
        run_settings_: Settings from the "HB Recipe Settings" component that specify
            how the recipe should be run. This can also be a text string of
            recipe settings.
        _run: Set to True to run the recipe and get results. This input can also be
            the integer "2" to run the recipe silently.

    Returns:
        report: Reports, errors, warnings, etc.
        avg_irr: The average irradiance in W/m2 for each sensor over the Wea time period.
        radiation: The cumulative radiation in kWh/m2 over the Wea time period.
"""

ghenv.Component.Name = 'HB Cumulative Radiation'
ghenv.Component.NickName = 'CumulativeRadiation'
ghenv.Component.Message = '1.3.0'
ghenv.Component.Category = 'HB-Radiance'
ghenv.Component.SubCategory = '3 :: Recipes'
ghenv.Component.AdditionalHelpFromDocStrings = '3'

try:
    from lbt_recipes.recipe import Recipe
except ImportError as e:
    raise ImportError('\nFailed to import lbt_recipes:\n\t{}'.format(e))

try:
    from ladybug_rhino.grasshopper import all_required_inputs, recipe_result
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component) and _run:
    # create the recipe and set the input arguments
    recipe = Recipe('cumulative-radiation')
    recipe.input_value_by_name('model', _model)
    recipe.input_value_by_name('wea', _wea)
    recipe.input_value_by_name('timestep', _timestep_)
    recipe.input_value_by_name('sky-density', _sky_density_)
    recipe.input_value_by_name('north', north_)
    recipe.input_value_by_name('grid-filter', grid_filter_)
    recipe.input_value_by_name('sensor-count', sensor_count_)
    recipe.input_value_by_name('radiance-parameters', radiance_par_)

    # run the recipe
    silent = True if _run > 1 else False
    project_folder = recipe.run(run_settings_, radiance_check=True, silent=silent)

    # load the results
    try:
        avg_irr = recipe_result(recipe.output_value_by_name('average-irradiance', project_folder))
        radiation = recipe_result(recipe.output_value_by_name('cumulative-radiation', project_folder))
    except Exception:
        raise Exception(recipe.failure_message(project_folder))



# Honeybee: A Plugin for Environmental Analysis (GPL)
# This file is part of Honeybee.
#
# Copyright (c) 2021, Ladybug Tools.
# You should have received a copy of the GNU General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""
Run an annual daylight study for a Honeybee model to compute hourly illuminance
for each sensor in a model's sensor grids.
_
This recipe uses an enhanced 2-phase method for daylight simulation which accurately
models direct sun by tracing rays from each sensor to the solar position at each
hour of the calculation.
_
The resulting illuminance is used to compute the following metrics:
_
* Daylight Autonomy (DA) - The percentage of occupied hours that each sensor
        recieves more than the illuminance threshold.
* Continuous Daylight Autonomy (cDA) - Similar to DA except that values below the
        illuminance threshold can still count partially towards the final percentage.
* Useful Daylight Illuminance (UDI) - The percentage of occupied hours that
        illuminace falls between minimum and maximum thresholds

-
    Args:
        _model: A Honeybee Model for which Annual Daylight will be simulated.
            Note that this model must have grids assigned to it.
        _wea: A Wea object produced from the Wea components that are under the Light
            Sources tab. This can also be the path to a .wea or a .epw file.
            Note that the Wea must have a timestep of 1 to be used with this
            recipe.
        north_: A number between -360 and 360 for the counterclockwise difference
            between the North and the positive Y-axis in degrees. This can
            also be Vector for the direction to North. (Default: 0).
        _thresholds_: A string to change the threshold for daylight autonomy and useful
            daylight illuminance. Valid keys are -t for daylight autonomy threshold,
            -lt for the lower threshold for useful daylight illuminance and
            -ut for the upper threshold. The order of the keys is not important
            and you can include one or all of them. For instance if you only want
            to change the upper threshold to 2000 lux you should use -ut 2000
            as the input. (Default: -t 300 -lt 100 -ut 3000).
        _schedule_: An annual occupancy schedule, either as a Ladybug Hourly Continuous
            Data Collection or a HB-Energy schedule object. This can also be the
            path to a CSV file with 8760 rows or the identifier of a schedule in
            the honeybee-energy schedule library. Any value in this schedule
            that is 0.1 or above will be considered occupied.
        grid_filter_: Text for a grid identifer or a pattern to filter the sensor grids of
            the model that are simulated. For instance, first_floor_* will simulate
            only the sensor grids that have an identifier that starts with
            first_floor_. By default, all grids in the model will be simulated.
        radiance_par_: Text for the radiance parameters to be used for ray
            tracing. (Default: -ab 2 -ad 5000 -lw 2e-05).
        run_settings_: Settings from the "HB Recipe Settings" component that specify
            how the recipe should be run. This can also be a text string of
            recipe settings.
        _run: Set to True to run the recipe and get results. This input can also be
            the integer "2" to run the recipe silently.

    Returns:
        report: Reports, errors, warnings, etc.
        results: Folder with raw result files (.ill) that contain illuminance matrices.
        DA: Daylight autonomy results in percent. DA is the percentage of occupied hours
            that each sensor recieves equal or more than the illuminance threshold.
            Each value is for a different sensor of the grid. These can be plugged
            into the "LB Spatial Heatmap" component along with meshes of the sensor
            grids to visualize results. These can also be connected to the "HB
            Spatial Daylight Autonomy" component to compute spatial daylight
            autonomy for each grid.
        cDA: Continuous daylight autonomy results in percent. cDA is similar to DA except
            that values below the illuminance threshold can still count partially
            towards the final percentage. Each value is for a different
            sensor of the grid. These can be plugged into the "LB Spatial Heatmap"
            component along with meshes of the sensor grids to visualize results.
        UDI: Useful daylight illuminance results in percent. UDI is the percentage of
            occupied hours that illuminace falls between minimum and maximum
            thresholds. Each value is for a different sensor of the grid. These
            can be plugged into the "LB Spatial Heatmap" component along with
            meshes of the sensor grids to visualize results.
        UDI_low: Results for the percent of time that is below the lower threshold
            of useful daylight illuminance in percent. Each value is for a different
            sensor of the grid. These can be plugged into the "LB Spatial Heatmap"
            component along with meshes of the sensor grids to visualize results.
        UDI_up: Results for the percent of time that is above the upper threshold
            of useful daylight illuminance in percent. Each value is for a different
            sensor of the grid. These can be plugged into the "LB Spatial Heatmap"
            component along with meshes of the sensor grids to visualize results.
"""

ghenv.Component.Name = 'HB Annual Daylight'
ghenv.Component.NickName = 'AnnualDaylight'
ghenv.Component.Message = '1.3.1'
ghenv.Component.Category = 'HB-Radiance'
ghenv.Component.SubCategory = '3 :: Recipes'
ghenv.Component.AdditionalHelpFromDocStrings = '1'

try:
    from lbt_recipes.recipe import Recipe
except ImportError as e:
    raise ImportError('\nFailed to import lbt_recipes:\n\t{}'.format(e))

try:
    from ladybug_rhino.grasshopper import all_required_inputs, recipe_result
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component) and _run:
    # create the recipe and set the input arguments
    recipe = Recipe('annual-daylight')
    recipe.input_value_by_name('model', _model)
    recipe.input_value_by_name('wea', _wea)
    recipe.input_value_by_name('north', north_)
    recipe.input_value_by_name('thresholds', _thresholds_)
    recipe.input_value_by_name('schedule', _schedule_)
    recipe.input_value_by_name('grid-filter', grid_filter_)
    recipe.input_value_by_name('radiance-parameters', radiance_par_)

    # run the recipe
    silent = True if _run > 1 else False
    project_folder = recipe.run(run_settings_, radiance_check=True, silent=silent)

    # load the results
    try:
        results = recipe_result(recipe.output_value_by_name('results', project_folder))
        DA = recipe_result(recipe.output_value_by_name('da', project_folder))
        cDA = recipe_result(recipe.output_value_by_name('cda', project_folder))
        UDI = recipe_result(recipe.output_value_by_name('udi', project_folder))
        UDI_low = recipe_result(recipe.output_value_by_name('udi-lower', project_folder))
        UDI_up = recipe_result(recipe.output_value_by_name('udi-upper', project_folder))
    except Exception:
        raise Exception(recipe.failure_message(project_folder))



# Honeybee: A Plugin for Environmental Analysis (GPL) started by Mostapha Sadeghipour Roudsari
# This file is part of Honeybee.
#
# You should have received a copy of the GNU General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""
Get cumulative radiation (or sum of illuminance) values over an annual irradiance
or daylight simulation.
_
The _hoys_ input can also be used to filter the data for a particular time period or
hour/timestep of the simulation.

-
    Args:
        _results: An list of annual Radiance result files from either the "HB Annual Daylight"
            or the "HB Annual Irradiance" component (containing the .ill files and
            the sun-up-hours.txt). This can also be just the path to the folder
            containing these result files.
        _hoys_: An optional numbers or list of numbers to select the hours of the year (HOYs)
            for which results will be computed. These HOYs can be obtained from the
            "LB Calculate HOY" or the "LB Analysis Period" components. If None, all
            hours of the results will be used.
        grid_filter_: The name of a grid or a pattern to filter the grids. For instance,
            first_floor_* will simulate only the sensor grids that have an
            identifier that starts with first_floor_. By default all the grids
            will be processed.

    Returns:
        report: Reports, errors, warnings, etc.
        values: In the case of an annual irradaince simulation, this is the cumulative
            radiation valules for each sensor in Wh/m2. For annual daylight, it is
            cumulative illuminance (lux-hours). These can be plugged into the "LB
            Spatial Heatmap" component along with meshes of the sensor
            grids to visualize results.
"""

ghenv.Component.Name = 'HB Annual Cumulative Values'
ghenv.Component.NickName = 'CumulValues'
ghenv.Component.Message = '1.3.0'
ghenv.Component.Category = 'HB-Radiance'
ghenv.Component.SubCategory = '4 :: Results'
ghenv.Component.AdditionalHelpFromDocStrings = '2'

import os

try:
    from honeybee_radiance.postprocess.annualdaylight import _process_input_folder
except ImportError as e:
    raise ImportError('\nFailed to import honeybee_radiance:\n\t{}'.format(e))

try:
    from ladybug_rhino.grasshopper import all_required_inputs, list_to_data_tree
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


def parse_sun_up_hours(sun_up_hours, hoys, timestep):
    """Parse the sun-up hours from the result file .txt file.

    Args:
        sun_up_hours: A list of integers for the sun-up hours.
        hoys: A list of 8760 * timestep values for the hoys to select. If an empty
            list is passed, None will be returned.
        timestep: Integer for the timestep of the analysis.
    """
    if len(hoys) != 0:
        schedule = [False] * (8760 * timestep)
        for hr in hoys:
            schedule[int(hr * timestep)] = True
        su_pattern = [schedule[int(h * timestep)] for h in sun_up_hours]
        return su_pattern


def cumulative_values(ill_file, su_pattern, timestep):
    """Compute average values for a given result file."""
    cumul_vals = []
    with open(ill_file) as results:
        if su_pattern is None:  # no HOY filter on results
            for pt_res in results:
                values = [float(r) for r in pt_res.split()]
                cumul_vals.append(sum(values) / timestep)
        else: 
            for pt_res in results:
                values = [float(r) for r, is_hoy in zip(pt_res.split(), su_pattern) if is_hoy]
                cumul_vals.append(sum(values) / timestep)
    return cumul_vals


if all_required_inputs(ghenv.Component):
    # get the relevant .ill files
    grid_filter_ = '*' if grid_filter_ is None else grid_filter_
    res_folder = os.path.dirname(_results[0]) if os.path.isfile(_results[0]) \
        else _results[0]
    grids, sun_up_hours = _process_input_folder(res_folder, grid_filter_)

    # extract the timestep if it exists
    timestep = 1
    tstep_file = os.path.join(res_folder, 'timestep.txt')
    if os.path.isfile(tstep_file):
        with open(tstep_file) as tf:
            timestep = int(tf.readline())

    # parse the sun-up-hours
    su_pattern = parse_sun_up_hours(sun_up_hours, _hoys_, timestep)

    # compute the average values
    values = []
    for grid_info in grids:
        ill_file = os.path.join(res_folder, '%s.ill' % grid_info['full_id'])
        cumul = cumulative_values(ill_file, su_pattern, timestep)
        values.append(cumul)
    values = list_to_data_tree(values)



#
# Honeybee: A Plugin for Environmental Analysis (GPL) started by Mostapha Sadeghipour Roudsari
# 
# This file is part of Honeybee.
# 
# Copyright (c) 2021, Ladybug Tools.
# Honeybee: A Plugin for Environmental Analysis (GPL)
# This file is part of Honeybee.
#
# Copyright (c) 2021, Ladybug Tools.
# You should have received a copy of the GNU General Public License
# along with Honeybee; If not, see <http://www.gnu.org/licenses/>.
# 
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""
Generate electric lighting schedules from annual daylight results, which can be
used to account for daylight controls in energy simulations.
_
Such controls will dim the lights in the energy simulation according to whether
the illuminance values at the sensor locations are at a target illuminance setpoint.
_
In addition to benefiting from the accuracy of Radiance, using this component has
several advantages over the "HB Apply Daylight Control" component under HB-Energy.
Notably, it can account for setups with multiple illuminance sensors.
_
This component expects the annual daylight simulation to be run with one sensor
grid per room in the model. If the sensor grids within the annual daylight
simulation are distrbuted over the entire floor of each room, the resulting
schedules will be idealized, where light dimming has been optimized to supply
the minimum illuminance setpoint everywhere in the room. Grids with one, two,
or more sensors can be used to model setups where fractions of each room are
controlled by different sensors.
-

    Args:
        _results: An list of annual Radiance result files from the "HB Annual Daylight"
            component (containing the .ill files and the sun-up-hours.txt).
            This can also be just the path to the folder containing these
            result files.
        _base_schedule_: A lighting schedule representing the usage of lights without
            any daylight controls. The values of this schedule will be multiplied
            by the hourly dimming fraction to yield the output lighting schedules.
            The format of this schedule can be a Ladybug Data Collection, a HB-Energy
            schedule object, or the identifier of a schedule in the HB-Energy
            schedule library. If None, a schedule from 9AM to 5PM on weekdays
            will be used.
        _ill_setpoint_: A number for the illuminance setpoint in lux beyond which
            electric lights are dimmed if there is sufficient daylight.
            Some common setpoints are listed below. (Default: 300 lux).
            -
            50 lux - Corridors and hallways.
            150 lux - Computer work spaces (screens provide illumination).
            300 lux - Paper work spaces (reading from surfaces that need illumination).
            500 lux - Retail spaces or museums illuminating merchandise/artifacts.
            1000 lux - Operating rooms and workshops where light is needed for safety.

        _min_power_in_: A number between 0 and 1 for the the lowest power the lighting
            system can dim down to, expressed as a fraction of maximum
            input power. (Default: 0.3).
        _min_light_out_: A number between 0 and 1 the lowest lighting output the lighting
            system can dim down to, expressed as a fraction of maximum light
            output. Note that setting this to 1 means lights aren't dimmed at
            all until the illuminance setpoint is reached. This can be used to
            approximate manual light-switching behaviour when used in conjunction
            with the off_at_min_ output below. (Default: 0.2).
        off_at_min_: Boolean to note whether lights should switch off completely when
            they get to the minimum power input. (Default: False).

    Returns:
        report: Reports, errors, warnings, etc.
        rooms: The input Rooms with simple daylight controls assigned to them.
"""

ghenv.Component.Name = 'HB Daylight Control Schedule'
ghenv.Component.NickName = 'DaylightSchedule'
ghenv.Component.Message = '1.3.0'
ghenv.Component.Category = 'HB-Radiance'
ghenv.Component.SubCategory = '4 :: Results'
ghenv.Component.AdditionalHelpFromDocStrings = '1'

import os

try:
    from ladybug.datacollection import BaseCollection
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))

try:
    from honeybee_radiance.postprocess.annualdaylight import \
        generate_default_schedule, _process_input_folder
except ImportError as e:
    raise ImportError('\nFailed to import honeybee_radiance:\n\t{}'.format(e))

try:
    from honeybee_energy.lib.schedules import schedule_by_identifier
    from honeybee_energy.lib.scheduletypelimits import schedule_type_limit_by_identifier
    from honeybee_energy.schedule.fixedinterval import ScheduleFixedInterval
except ImportError as e:
    raise ImportError('\nFailed to import honeybee_energy:\n\t{}'.format(e))

try:
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


def diminng_from_ill(ill_val, ill_setpt, min_pow, min_light, off_at_min):
    """Compute the dimming fraction from an illuminance value."""
    if ill_val > ill_setpt:  # dimmed all of the way
        return 0 if off_at_min else min_pow
    elif ill_val <= min_light:  # not dimmed at all
        return 1
    else:  # partially dimmed
        fract_dim = (ill_setpt - ill_val) / (ill_setpt - min_light)
        return fract_dim + ((1 - fract_dim) * min_pow)


def file_to_dimming_fraction(ill_file, su_pattern, setpt, m_pow, m_lgt, off_m):
    """Compute hourly dimming fractions for a given result file."""
    # get a base schedule of dimming fractions for the sun-up hours
    su_values = [0] * len(su_pattern)
    sensor_count = 0
    with open(ill_file) as results:
        for pt_res in results:
            sensor_count += 1
            for i, val in enumerate(pt_res.split()):
                su_values[i] += diminng_from_ill(float(val), setpt, m_pow, m_lgt, off_m)
    su_values = [val / sensor_count for val in su_values]

    # account for the hours where the sun is not up
    dim_fract = [1] * 8760
    for val, hr in zip(su_values, su_pattern):
        dim_fract[hr] = float(val)
    return dim_fract


if all_required_inputs(ghenv.Component):
    # set default values for all controls
    _ill_setpoint_ = 300 if _ill_setpoint_ is None else _ill_setpoint_
    _min_power_in_ = 0.3 if _min_power_in_ is None else _min_power_in_
    _min_light_out_ = 0.2 if _min_light_out_ is None else _min_light_out_
    off_at_min_ = False if off_at_min_ is None else off_at_min_

    # process the base schedule input into a list of values
    if _base_schedule_ is None:
        schedule = generate_default_schedule()
    elif isinstance(_base_schedule_, BaseCollection):
        schedule = _base_schedule_.values
    elif isinstance(_base_schedule_, str):
        schedule = schedule_by_identifier(_base_schedule_).values()
    else:  # assume that it is a honeybee schedule object
        try:
            schedule = _base_schedule_.values()
        except TypeError:  # it's probably a ScheduleFixedInterval
            schedule = _base_schedule_.values

    # get the relevant .ill files
    res_folder = os.path.dirname(_results[0]) if os.path.isfile(_results[0]) \
        else _results[0]
    grids, sun_up_hours = _process_input_folder(res_folder, '*')
    sun_up_hours = [int(h) for h in sun_up_hours]

    # get the dimming fractions for each sensor grid from the .ill files
    dim_fracts = []
    for grid_info in grids:
        ill_file = os.path.join(res_folder, '%s.ill' % grid_info['full_id'])
        fract_list = file_to_dimming_fraction(
            ill_file, sun_up_hours, _ill_setpoint_, _min_power_in_,
            _min_light_out_, off_at_min_
        )
        dim_fracts.append(fract_list)

    # create the schedule by combining the base schedule with the dimming fraction
    type_limit = schedule_type_limit_by_identifier('Fractional')
    schedules = []
    for grid_info, dim_fract in zip(grids, dim_fracts):
        sch_vals = [b_val * d_val for b_val, d_val in zip(schedule, dim_fract)]
        sch_id = '{} Daylight Control'.format(grid_info['full_id'])
        schedules.append(ScheduleFixedInterval(sch_id, sch_vals, type_limit))
